
//		ProcTree.cpp - Display a process tree


#define UNICODE								1
#define _UNICODE							1

#include <stdio.h>
#include <tchar.h>

#include <windows.h>
#include <TlHelp32.h>
#include <Psapi.h>


#include "PlugIn.h"
#include "TakeCmd.h"


#define PLUGIN_NAME							L"ProcTree"

#define VER_MAJOR							1
#define VER_MINOR							1
#define VER_BUILD							1
#define VER_PATCH							0

#ifdef _WIN64
#define PLATFORM							L"x64"
#else
#define PLATFORM							L"x86"
#endif

#define MAX_ARG_LEN							8192
#define SHORT_BUF_LEN						1024
#define MAX_FILENAME_LEN					32768


#define ERRHANDLE							GetStdHandle( STD_ERROR_HANDLE )


#define PROCTREE_FLAG_ASCII					0x000001
#define PROCTREE_FLAG_ASCII_ELLIPSES		0x000002
#define PROCTREE_FLAG_DECIMAL				0x000004
#define PROCTREE_FLAG_NO_SERVICE_NAMES		0x000008
#define PROCTREE_FLAG_NO_WINDOW_TITLES		0x000010
#define PROCTREE_FLAG_NO_TREE_GRAPHIC		0x000020
#define ANCESTRY_FLAG_LINE_NUMBERS			0x000020
#define PROCTREE_FLAG_NO_DISPLAY_STRINGS	0x000040
#define PROCTREE_FLAG_SHOW_THREAD_COUNT		0x000080
#define PROCTREE_FLAG_SHOW_PARENT			0x000100
#define PROCTREE_FLAG_SHOW_PRIORITY			0x000200
#define PROCTREE_FLAG_SHOW_TIMES			0x000400
#define PROCTREE_FLAG_INCLUDE_DATES			0x000800
#define PROCTREE_FLAG_INCLUDE_SECONDS		0x001000
#define PROCTREE_FLAG_INCLUDE_MS			0x002000
#define PROCTREE_FLAG_SHOW_EXTRA_INFO		0x004000
#define PROCTREE_FLAG_GET_EXTRA_INFO		0x008000
#define PROCTREE_FLAG_FIND_SERVICES			0x010000
#define PROCTREE_FLAG_SHOW_EXTENSIONS		0x020000
#define PROCTREE_FLAG_SHOW_SUMMARY			0x040000
#define PROCTREE_FLAG_SHOW_MEMORY_INFO		0x080000
#define PROCTREE_FLAG_SHOW_MEMORY_BYTES		0x100000
#define PROCTREE_FLAG_SHOW_MEMORY_FRACTIONS	0x200000
#define PROCTREE_FLAG_SHOW_FILENAMES		0x400000
#define ANCESTRY_FLAG_REVERSE_ORDER			0x800000



#define PROCTREE_WIDTH_MIN					40
#define PROCTREE_WIDTH_MAX					1020
#define PROCTREE_WIDTH_DEFAULT				100

#define PROCTREE_SORTKEY_LEN				16

#define PROCTREE_ERROR_PID_NOT_FOUND		42
#define PROCTREE_ERROR_NO_MATCH				43
#define PROCTREE_ERROR_SERVICES_NOT_FOUND	44



OSVERSIONINFOEXW OSVerInfo;


typedef BOOL ( WINAPI _QueryFullProcessImageName ) ( HANDLE Process, DWORD Flags, LPWSTR ExeName, PDWORD Size );
_QueryFullProcessImageName *QueryFullProcessImageNameProc	= NULL;



const wchar_t PluginFeatures[] = L"ancestry,proctree,@ancestry";



const wchar_t ProcTreeHelpText[]	= L"List descendants of a given process.\n\
\n\
PROCTREE /A /C /D /E /F /H: /I /K /M /N: /O: /SVC /TC /T: /W: root  \r\n\
\n\
\t/A\tASCII characters\n\
\t/C\tshow priorities\n\
\t/D\tdecimal PIDs\n\
\t/E\tstart from Explorer\n\
\t/F\tshow complete filenames\n\
\t/H:\thighlight matching strings\n\
\t/I\tshow extra info:  \u2191 elevated, \u00bb 32-bit\n\
\t/K\tshow thread counts\n\
\t/M\tshow memory usage\n\
\t/N:\tdisable features:  CDESTWX  \t/N? for help\n\
\t/O:\tsort processes:    -CDEKMNPTW \t/O? for help\n\
\t/SVC\tstart from services root\n\
\t/TC\tstart from Take Command\n\
\t/T:\tprocess start times:  DSMF    \t/T? for help\n\
\t/W:\tspecify line width\n\
\n\
\u201cRoot\u201d may be a PID in decimal, or hexadecimal with a leading 0x;\n\
or a wildspec matching the program\u2019s executable name, window title,  \n\
or service name.  If omitted, it defaults to the current shell.\n";


const wchar_t ProcTreeOptNHelpText[]	= L"PROCTREE /N:  \u2014 disable features\n\
\n\
\tC\tno highlight\n\
\tD\tdo not show display names\n\
\tE\tASCII periods instead of Unicode ellipses\n\
\tS\tdo not collect service names\n\
\tT\tdo not display the tree graphic\n\
\tW\tdo not collect window titles\n\
\tX\tdo not hide extensions\n";


const wchar_t ProcTreeOptOHelpText[]	= L"PROCTREE /O:  \u2014 sort sibling processes\n\
\n\
\tC\tby priority (ascending)\n\
\tD\tby display name\n\
\tE\televated processes first\n\
\tK\tby thread count (ascending)\n\
\tM\tby memory usage (ascending)\n\
\tN\tby executable name\n\
\tP\tby process ID (ascending)\n\
\tT\tby creation time (oldest first)\n\
\tW\t32-bit processes first\n\
\t-\treverses the following key\n";


const wchar_t ProcTreeOptTHelpText[]	= L"PROCTREE /T:  \u2014 show start times\n\
\n\
\tD\tdate\n\
\tS\tseconds\n\
\tM\tmilliseconds\n\
\tF\tall of the above\n";


const wchar_t AncestryHelpText[]		= L"List ancestors of a given process.\n\
\n\
ANCESTRY /C /D /F /H: /I /K /M /N: /R /T: /W: process  \r\n\
\n\
\t/C\tshow priorities\n\
\t/D\tdecimal PIDs\n\
\t/F\tshow complete filenames\n\
\t/H:\thighlight matching strings\n\
\t/I\tshow extra info:  \u2191 elevated, \u00bb 32-bit\n\
\t/K\tshow thread counts\n\
\t/M\tshow memory usage\n\
\t/N:\tdisable features:  CDESWX  \t/N? for help\n\
\t/R\treverse order\n\
\t/T:\tprocess start times:  DSMF    \t/T? for help\n\
\t/W:\tspecify line width\n\
\n\
\u201cProcess\u201d may be a PID in decimal, or hexadecimal with a leading 0x;  \n\
or a wildspec matching the program\u2019s executable name, window title,\n\
or service name.  If omitted, it defaults to the current shell.\n";


const wchar_t AncestryOptNHelpText[]	= L"ANCESTRY /N:  \u2014 disable features\n\
\n\
\tC\tno highlight\n\
\tD\tdo not show display names\n\
\tE\tASCII periods instead of Unicode ellipses\n\
\tS\tdo not collect service names\n\
\tW\tdo not collect window titles\n\
\tX\tdo not hide extensions\n";


const wchar_t AncestryOptTHelpText[]	= L"ANCESTRY /T:  \u2014 show start times\n\
\n\
\tD\tdate\n\
\tS\tseconds\n\
\tM\tmilliseconds\n\
\tF\tall of the above\n";


const wchar_t F_AncestryHelpText[]		= L"Returns ancestors of a given process.\n\
\n\
@ANCESTRY[pid,flags] \r\n\
\n\
\tpid\tthe process ID, decimal or hex  \n\
\tflags\t\u2000\u20000:  return PIDS in hexadecimal\n\
\t\t\u2000\u20001:  hexadecimal, no leading 0x  \n\
\t\t\u2000\u20002:  decimal\n\
\t\t\u2000\u20003:  executable names\n\
\t\t\u200064:  reverse order\n\
\t\t128:  lowercase\n\
\n\
If pid is not specified, it defaults to the current shell.  \n";



const wchar_t TreeGraf[8][4]	= {
	L"\u251c\u2500\u2500", L"+--",		//	branch graphic (not the last sibling)
	L"\u2514\u2500\u2500", L"`--",		//	branch graphic (last sibling)
	L"\u2502  ", L"|  ",				//	append to prefix (parent was not the last sibling)
	L"   ", L"   "						//	append to prefix (parent was the last sibling)
};



const wchar_t PriorityStr[7][3]		= { L"-", L"I", L"BN", L"N", L"AN", L"HI", L"RT" };



int DisplayErrorHeader()
{
	wchar_t shortbuf[MAX_ARG_LEN] = L"%_batch";

	Qprintf( ERRHANDLE, L"%s plugin: ", PLUGIN_NAME );

	ExpandVariables( shortbuf, 1 );
	if ( shortbuf[0] == '0' )
		return 0;

	wcscpy_s( shortbuf, MAX_ARG_LEN, L"%_batchname [%_batchline]  " );
	ExpandVariables( shortbuf, 1 );
	Qprintf( ERRHANDLE, shortbuf );

	return 0;
}



BOOL IsColonOrEquals( wchar_t ch )
{
	return ( ( ch == ':' ) || ( ch == '=' ) );
}



int ParseInt( LPTSTR inString, unsigned int *outValue, wchar_t **EndChar )
{
	if ( !inString || !outValue )
		return -1;

	unsigned int value	= 0;
	wchar_t *sep1		= NULL;

	_set_errno( 0 );

	if ( ( inString[0] == '0' ) && ( towupper( inString[1] ) == 'X' ) && ( iswxdigit( inString[2] ) ) ) {
		//  parse hexadecimal

		value = wcstoul( inString + 2, &sep1, 16 );

		if ( errno )
			return -1;
		else if ( !EndChar && ( *sep1 != '\0' ) )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}
	else if ( iswdigit( inString[0] ) ) {
		//  parse unsigned

		value = wcstoul( inString, &sep1, 10 );

		if ( errno )
			return -1;
		else if ( !EndChar && ( *sep1 != '\0' ) )
			return -1;

		*outValue = value;

		if ( EndChar )
			*EndChar  = sep1;

		return 0;
	}

	return -1;
}



#include "ParseArgs.cpp"
#include "conlist.cpp"
#include "Ancestry.cpp"



BOOL GetVersion2( OSVERSIONINFOEXW *VerInfo )
{
	if ( !VerInfo )
		return -1;


	typedef NTSTATUS ( WINAPI _RtlGetVersionProc ) ( OSVERSIONINFOEXW *VerInfo );
	HMODULE NtDllHandle						= GetModuleHandle( L"NtDll.dll" );
	_RtlGetVersionProc *RtlGetVersionProc	= NULL;


	memset( VerInfo, 0x00, sizeof( OSVERSIONINFOEXW ) );
	VerInfo->dwOSVersionInfoSize = sizeof( OSVERSIONINFOEXW );

	if ( NtDllHandle )
		RtlGetVersionProc = (_RtlGetVersionProc*) GetProcAddress( NtDllHandle, "RtlGetVersion" );

	if ( RtlGetVersionProc ) {
		RtlGetVersionProc( VerInfo );
		if ( VerInfo->dwMajorVersion >= 6 )
			return TRUE;
	}

	memset( VerInfo, 0x00, sizeof( OSVERSIONINFOEXW ) );
	VerInfo->dwOSVersionInfoSize = sizeof( OSVERSIONINFOEXW );

	return ( GetVersionEx( (OSVERSIONINFOW*) VerInfo ) != 0 );
}



int GetErrorText( int err, LPTSTR outString, unsigned int Flags = 0 )
{
	if ( outString == NULL )
		return -1;


	HANDLE WinINetHandle		= NULL;
	DWORD FM_Flags				= FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS | FORMAT_MESSAGE_MAX_WIDTH_MASK;

	
	if ( ( err >= 12000 ) && ( err <= 12199 ) ) {
		WinINetHandle = GetModuleHandle( L"wininet.dll" );
		if ( WinINetHandle )
			FM_Flags |= FORMAT_MESSAGE_FROM_HMODULE;
	}


	if ( err == -1 )
		wcscpy_s( outString, SHORT_BUF_LEN, L"Unknown error" );
	else if ( !FormatMessage( FM_Flags, WinINetHandle, err, 0, outString, SHORT_BUF_LEN, NULL ) ) {
		if ( Flags & 1 )
			outString[0] = '\0';
		else
			wsprintf( outString, L"Error %d", err );
	}

	size_t l = wcslen( outString );

	if ( l )
		while ( iswspace( outString[l - 1] ) ) {
			l--;
			outString[l] = '\0';
		}

	return 0;
}



int GetLastErrorText( LPTSTR outString )
{
	return GetErrorText( GetLastError(), outString );
}



//	Deletes all items in a linked list, and zeros the first-item pointer:

int FreeLinkedList( void **FirstItem )
{
	if ( FirstItem == NULL )
		return -1;

	size_t *This = (size_t*) *FirstItem, Next = NULL;

	while ( This ) {
		Next = *This;
		free( This );
		This = (size_t*) Next;
	}

	*FirstItem = NULL;
	return 0;
}



int MatchWildcardList( LPCTSTR Filename, LPCTSTR WildSpecs, int Ext, int Brackets, int AssumeDotStar )
{
	if ( !Filename || !WildSpecs )
		return 0;


	wchar_t temp[MAX_ARG_LEN]			= L"";

	const wchar_t *ch		= WildSpecs;
	int j					= 0;
	int k					= 1;


	while ( *ch ) {

		if ( *ch == '|' ) {
			k++;

			if ( temp[0] ) {
				if ( AssumeDotStar && !wcschr( temp, '.' ) )
					wcscat_s( temp, MAX_ARG_LEN, L".*" );

				if ( ( temp[0] == '!' ) && temp[1] ) {
					if ( WildcardComparison( temp + 1, (LPTSTR) Filename, Ext, Brackets ) )
						return 1;
				}
				else if ( !WildcardComparison( temp, (LPTSTR) Filename, Ext, Brackets ) )
					return 1;
			}

			j = 0;
			temp[j] = '\0';
		}
		else if ( j < ( MAX_ARG_LEN - 1 ) ) {
			temp[j++] = *ch;
			temp[j] = '\0';
		}

		ch++;
	}

	k++;

	if ( temp[0] ) {
		if ( AssumeDotStar && !wcschr( temp, '.' ) )
			wcscat_s( temp, MAX_ARG_LEN, L".*" );

		if ( ( temp[0] == '!' ) && temp[1] ) {
			if ( WildcardComparison( temp + 1, (LPTSTR) Filename, Ext, Brackets ) )
				return 1;
		}
		else if ( !WildcardComparison( temp, (LPTSTR) Filename, Ext, Brackets ) )
			return k;
	}

	return 0;
}



BOOL APIENTRY DllMain( HANDLE hModule, DWORD  dwReason, LPVOID lpReserved )
{
	switch( dwReason )	{
		case DLL_PROCESS_ATTACH:
		case DLL_THREAD_ATTACH:
		case DLL_THREAD_DETACH:
		case DLL_PROCESS_DETACH:
			break;
	}
	return TRUE;
}



DLLExports LPPLUGININFO WINAPI GetPluginInfo( void )
{
	static PLUGININFO piInfo;

	piInfo.pszDll			= (LPTSTR) PLUGIN_NAME;
	piInfo.pszAuthor		= (LPTSTR) L"Charles Dye";
	piInfo.pszEmail			= (LPTSTR) L"cdye@unm.edu";
	piInfo.pszWWW			= (LPTSTR) L"http://charlesdye.net/plugins/proctree.html";
	piInfo.pszDescription	= (LPTSTR) L"Display a process tree.";
	piInfo.pszFunctions		= (LPTSTR) PluginFeatures;
	piInfo.nMajor			= VER_MAJOR;
	piInfo.nMinor			= VER_MINOR;
	piInfo.nBuild			= VER_BUILD;

	return &piInfo;
}



DLLExports BOOL WINAPI InitializePlugin( void )
{
	GetVersion2( &OSVerInfo );

	QueryFullProcessImageNameProc = (_QueryFullProcessImageName*) GetProcAddress( GetModuleHandle( L"Kernel32.dll" ), "QueryFullProcessImageNameW" );


	wchar_t shortbuf[SHORT_BUF_LEN] = L"%[_transient]%[_pipe]%[noloadmsg]";
	ExpandVariables( shortbuf, 1 );

	if ( wcscmp( shortbuf, L"00" ) == 0 )
		Printf( L"%s v%u.%u.%u loaded.\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD );

	return 0;
}



DLLExports BOOL WINAPI ShutdownPlugin( BOOL bEndProcess )
{

#ifdef NEWHELP_VERSION
	CloseHelpWindow();
#endif

	return 0;
}


//	--------------------------  Used to collect service names:


struct _PidSvcName {
	_PidSvcName *Next;
	DWORD Pid;
	unsigned int Count;
	wchar_t ServiceNames[SHORT_BUF_LEN];
};



_PidSvcName *NewPidSvcName( DWORD Pid, LPCTSTR ServiceName )
{
	if ( !Pid || !ServiceName )
		return NULL;

	if ( ServiceName[0] == '\0' )
		return NULL;


	_PidSvcName *New	= (_PidSvcName*) malloc( sizeof( _PidSvcName ) );
	if ( !New )
		return NULL;

	New->Next	= NULL;
	New->Pid	= Pid;
	New->Count	= 1;
	wcscpy_s( New->ServiceNames, SHORT_BUF_LEN, ServiceName );

	return New;
}



int AddPidSvcName( DWORD Pid, LPTSTR ServiceName, _PidSvcName **First )
{
	if ( !First || !Pid || !ServiceName )
		return -1;

	if ( ServiceName[0] == '\0' )
		return -1;


	_PidSvcName *New		= NULL;
	_PidSvcName *This		= NULL;
	_PidSvcName *Last		= NULL;


	if ( *First == NULL ) {
		New = NewPidSvcName( Pid, ServiceName );
		if ( !New )
			return ERROR_NOT_ENOUGH_MEMORY;

		*First = New;
		return 0;
	}

	This = *First;

	while ( This ) {
		Last = This;

		if ( This->Pid == Pid ) {
			This->Count += 1;
			wcscat_s( This->ServiceNames, SHORT_BUF_LEN, L"; " );
			wcscat_s( This->ServiceNames, SHORT_BUF_LEN, ServiceName );
			return 0;
		}

		This = This->Next;
	}

	New = NewPidSvcName( Pid, ServiceName );
	if ( !New )
		return ERROR_NOT_ENOUGH_MEMORY;

	Last->Next	= New;
	return 0;
}



LPTSTR SvcNamesForPid( DWORD Pid, _PidSvcName *First )
{
	if ( !First || !Pid )
		return NULL;


	_PidSvcName *This		= First;


	while ( This ) {
		if ( This->Pid == Pid )
			return This->ServiceNames;

		This = This->Next;
	}

	return NULL;
}



int CollectPidServiceNames( _PidSvcName **First )
{
	if ( !First )
		return -1;


	SC_HANDLE SCHandle						= NULL;
	ENUM_SERVICE_STATUS_PROCESS *Services	= NULL;
	DWORD BytesNeeded						= 0;
	DWORD ServicesReturned					= 0;
	int Count								= 0;


	FreeLinkedList( (void**) First );


	SCHandle = OpenSCManager( NULL, NULL, SC_MANAGER_CONNECT | SC_MANAGER_ENUMERATE_SERVICE );
	if ( !SCHandle )
		return 0;

	EnumServicesStatusEx( SCHandle, SC_ENUM_PROCESS_INFO, SERVICE_DRIVER | SERVICE_WIN32, SERVICE_STATE_ALL, NULL, 0, &BytesNeeded, &ServicesReturned, NULL, NULL );

	if ( BytesNeeded ) {
		BytesNeeded += ( 4 * sizeof( ENUM_SERVICE_STATUS_PROCESS ) );

		Services = (ENUM_SERVICE_STATUS_PROCESS*) malloc( BytesNeeded );

		if ( Services ) {
			if ( EnumServicesStatusEx( SCHandle, SC_ENUM_PROCESS_INFO, SERVICE_DRIVER | SERVICE_WIN32, SERVICE_STATE_ALL, (BYTE*) Services, BytesNeeded, &BytesNeeded, &ServicesReturned, NULL, NULL ) ) {

				for ( unsigned int i = 0; i < ServicesReturned; i++ ) {
					//	wprintf( L"   * Service %u: 0x%04X  '%s'\n", i, Services[i].ServiceStatusProcess.dwProcessId, Services[i].lpDisplayName );

					if ( AddPidSvcName( Services[i].ServiceStatusProcess.dwProcessId, Services[i].lpDisplayName, First ) == 0 )
						Count++;
				}
			}

			free( Services );
		}
	}

	CloseServiceHandle( SCHandle);
	return Count;
}



//	--------------------------	Used to collect window titles:


struct _PidWinTitle {
	_PidWinTitle *Next;
	_PidWinTitle *Prev;
	DWORD Pid;
	unsigned int WinFlags;
	wchar_t Title[1];
};



_PidWinTitle *NewPidWinTitle( DWORD Pid, LPCTSTR Title, unsigned int WinFlags )
{
	if ( !Pid || !Title )
		return NULL;

	if ( Title[0] == '\0' )
		return NULL;


	_PidWinTitle *New		= (_PidWinTitle*) malloc( sizeof( _PidWinTitle ) + ( wcslen( Title ) * sizeof( wchar_t ) ) );
	if ( !New )
		return NULL;

	New->Next		= NULL;
	New->Prev		= NULL;
	New->Pid		= Pid;
	New->WinFlags	= WinFlags;
	wcscpy_s( New->Title, wcslen( Title ) + 1, Title );

	return New;
}



_PidWinTitle *FindPidWinTitle( DWORD Pid, _PidWinTitle *First )
{
	if ( !Pid || !First )
		return NULL;


	_PidWinTitle *This		= First;


	while ( This ) {
		if ( This->Pid == Pid )
			return This;

		This = This->Next;
	}

	return NULL;
}



LPTSTR PidWinTitle( DWORD Pid, _PidWinTitle *First )
{
	if ( !Pid || !First )
		return NULL;


	_PidWinTitle *This		= First;


	while ( This ) {
		if ( This->Pid == Pid )
			return This->Title;

		This = This->Next;
	}

	return NULL;
}



BOOL CALLBACK CollectPidWinTitlesProc$( HWND hWnd, LPARAM First )
{
	wchar_t Title[SHORT_BUF_LEN]		= L"";
	wchar_t ClassName[SHORT_BUF_LEN]	= L"";

	_PidWinTitle **FirstWin			= (_PidWinTitle**) First;
	DWORD Pid						= 0;
	LONG_PTR ExStyles				= 0;
	unsigned int WinFlags			= 0;
	BOOL Replace					= FALSE;

	_PidWinTitle *Old				= NULL;
	_PidWinTitle *New				= NULL;
	_PidWinTitle *This				= NULL;



	GetWindowThreadProcessId( hWnd, &Pid );
	GetWindowText( hWnd, Title, SHORT_BUF_LEN );
	RealGetWindowClass( hWnd, ClassName, SHORT_BUF_LEN );
	ExStyles = GetWindowLongPtrW( hWnd, GWL_EXSTYLE );


	if ( !Pid || ( wcslen( Title ) < 2 ) )
		return TRUE;

	if ( !wcscmp( ClassName, L"MSCTFIME UI" ) || !wcscmp( ClassName, L"IME" ) )
		return TRUE;

	if ( ExStyles & WS_EX_APPWINDOW )
		WinFlags |= 0x0010;

	if ( wcscmp( ClassName, L"#32770" ) )
		WinFlags |= 0x0004;

	if ( IsWindowVisible( hWnd ) )
		WinFlags |= 0x0001;


	//	wprintf( L"   * Found window 0x%08X:  Pid = 0x%04X  Title = '%s'\n", hWnd, Pid, Title );

	Old = FindPidWinTitle( Pid, *FirstWin );

	if ( Old ) {
		if ( WinFlags > Old->WinFlags )
			Replace = TRUE;
		else
			return TRUE;
	}


	New = NewPidWinTitle( Pid, Title, WinFlags );
	if ( !New )
		return FALSE;


	if ( *FirstWin == NULL ) {
		//	wprintf( L"   * Start new list with 0x%08X '%s':\n", hWnd, Title );

		*FirstWin = New;
	}
	else if ( Replace ) {
		//	wprintf( L"   * Replace '%s' (%04x) with 0x%08X '%s' (%04x):\n", Old->Title, Old->WinFlags, hWnd, Title, WinFlags );

		if ( Old->Prev ) {
			Old->Prev->Next = New;
			
			if ( Old->Next )
				Old->Next->Prev = New;

			New->Prev = Old->Prev;
			New->Next = Old->Next;
		}
		else {
			*FirstWin = New;

			if ( Old->Next )
				Old->Next->Prev = New;

			New->Next = Old->Next;
		}

		free( Old );
	}
	else {
		//	wprintf( L"   * Add 0x%08X '%s' to end of list:\n", hWnd, Title );

		This = *FirstWin;

		while ( This->Next )
			This = This->Next;

		This->Next = New;
		New->Prev = This;
	}

	return TRUE;
}



int CollectPidWinTitles( _PidWinTitle **First )
{
	if ( !First )
		return -1;


	int Count				= 0;


	FreeLinkedList( (void**) First );


	EnumWindows( CollectPidWinTitlesProc$, (LPARAM) First );

	return Count;
}


//	--------------------------



struct _ProcInfo {
	PROCESSENTRY32 pe;
	PROCESS_MEMORY_COUNTERS mi;
	FILETIME StartTime;
	LPCTSTR Title;
	BOOL Elevated;
	BOOL Wow64;
	DWORD Priority;
};



BOOL ProcessElevated( HANDLE ProcessHandle )
{
	if ( !ProcessHandle )
		return FALSE;


	HANDLE ProcessToken			= NULL;
	TOKEN_ELEVATION Elevated	= { 0 };
	DWORD ReturnSize			= 0;
	BOOL rv						= FALSE;


	if ( OpenProcessToken( ProcessHandle, TOKEN_QUERY, &ProcessToken ) ) {
		if ( GetTokenInformation( ProcessToken, TokenElevation, &Elevated, sizeof( TOKEN_ELEVATION ), &ReturnSize ) )
			rv = ( Elevated.TokenIsElevated != 0 );

		CloseHandle( ProcessToken );
	}

	return rv;
}



DWORD PriorityLevel( DWORD PriorityClass )
{
	switch ( PriorityClass ) {

	case REALTIME_PRIORITY_CLASS :
		return 6;

	case HIGH_PRIORITY_CLASS :
		return 5;

	case ABOVE_NORMAL_PRIORITY_CLASS :
		return 4;

	case NORMAL_PRIORITY_CLASS :
		return 3;

	case BELOW_NORMAL_PRIORITY_CLASS :
		return 2;

	case IDLE_PRIORITY_CLASS :
		return 1;

	default :
		return 0;
	}
}



BOOL ProcTree_ProcessIsNewer( DWORD ThisPid, DWORD ThatPid )
{
	if ( !ThisPid || !ThatPid )
		return FALSE;


	DWORD ProcAccess			= ( OSVerInfo.dwMajorVersion < 6 ) ? PROCESS_QUERY_INFORMATION : PROCESS_QUERY_LIMITED_INFORMATION;
	HANDLE ThisHandle			= NULL;
	HANDLE ThatHandle			= NULL;
	FILETIME ThisCreateTime		= { 0, 0 };
	FILETIME ThatCreateTime		= { 0, 0 };
	FILETIME ExitTime			= { 0, 0 };
	FILETIME KernelTime			= { 0, 0 };
	FILETIME UserTime			= { 0, 0 };
	BOOL rv						= FALSE;


	ThisHandle	= OpenProcess( ProcAccess, FALSE, ThisPid );
	ThatHandle	= OpenProcess( ProcAccess, FALSE, ThatPid );

	if ( ThisHandle && ThatHandle ) {
		if ( GetProcessTimes( ThisHandle, &ThisCreateTime, &ExitTime, &KernelTime, &UserTime ) && GetProcessTimes( ThatHandle, &ThatCreateTime, &ExitTime, &KernelTime, &UserTime ) ) {
			if ( ThisCreateTime.dwHighDateTime > ThatCreateTime.dwHighDateTime )
				rv = TRUE;
			else if ( ( ThisCreateTime.dwHighDateTime == ThatCreateTime.dwHighDateTime ) && ( ThisCreateTime.dwLowDateTime > ThatCreateTime.dwLowDateTime ) )
				rv = TRUE;
		}
	}


	CloseHandle( ThisHandle );
	CloseHandle( ThatHandle );
	return rv;
}



_ProcInfo *FindProcessEntry( DWORD Pid, _ProcInfo *ProcessList, unsigned int NumProcesses )
{
	if ( !ProcessList || !NumProcesses )
		return NULL;

	for ( unsigned int i = 0; i < NumProcesses; i++ )
		if ( ProcessList[i].pe.th32ProcessID == Pid )
			return ProcessList + i;

	return NULL;
}



unsigned int CountProcessEntries( HANDLE Snapshot )
{
	PROCESSENTRY32 ProcessInfo;
	BOOL Continue				= FALSE;
	unsigned int k				= 0;


	memset( &ProcessInfo, 0x00, sizeof( PROCESSENTRY32 ) );
	ProcessInfo.dwSize = sizeof( PROCESSENTRY32 );


	Continue = Process32First( Snapshot, &ProcessInfo );

	while ( Continue ) {
		k++;

		memset( &ProcessInfo, 0x00, sizeof( PROCESSENTRY32 ) );
		ProcessInfo.dwSize = sizeof( PROCESSENTRY32 );
		Continue = Process32Next( Snapshot, &ProcessInfo );
	}

	if ( GetLastError() != ERROR_NO_MORE_FILES )
		return 0;

	return k;
}



int DWordCmp( DWORD d1, DWORD d2 )
{
	if ( d1 < d2 )
		return -1;
	else if ( d1 > d2 )
		return 1;
	else
		return 0;
}



int ProcTreeCompare$( void *SortKeys, const void *Proc1, const void *Proc2 )
{
	const wchar_t *key		= (LPCTSTR) SortKeys;
	const _ProcInfo **a1	= (const _ProcInfo**) Proc1;
	const _ProcInfo **a2	= (const _ProcInfo**) Proc2;
	const _ProcInfo *p1		= *a1;
	const _ProcInfo *p2		= *a2;

	int rv					= 0;


	while ( *key ) {

		switch ( towupper( *key ) ) {

		case 'N' :

			rv = _wcsicmp( p1->pe.szExeFile, p2->pe.szExeFile );
			break;

		case 'P' :

			rv = DWordCmp( p1->pe.th32ProcessID, p2->pe.th32ProcessID );
			break;

		case 'T' :

			rv = DWordCmp( p1->StartTime.dwHighDateTime, p2->StartTime.dwHighDateTime );

			if ( !rv )
				rv = DWordCmp( p1->StartTime.dwLowDateTime, p2->StartTime.dwLowDateTime );

			break;

		case 'K' :

			rv = DWordCmp( p1->pe.cntThreads, p2->pe.cntThreads );
			break;

		case 'E' :

			rv = DWordCmp( p2->Elevated, p1->Elevated );
			break;

		case 'W' :

			rv = DWordCmp( p2->Wow64, p1->Wow64 );			
			break;

		case 'D' :

			if ( ( p1->Title == NULL ) && p2->Title )
				rv = -1;
			else if ( p1->Title && ( p2->Title == NULL ) )
				rv = 1;
			else if ( p1->Title && p2->Title )
				rv = _wcsicmp( p1->Title, p2->Title );

			break;

		case 'C':

			rv = DWordCmp( p1->Priority, p2->Priority );
			break;

		case 'M' :

			if ( p1->mi.WorkingSetSize < p2->mi.WorkingSetSize )
				rv = -1;
			else if ( p1->mi.WorkingSetSize > p2->mi.WorkingSetSize )
				rv = 1;
		}

		if ( iswupper( *key ) )
			rv = -rv;

		if ( rv )
			return rv;

		key++;
	}


	return DWordCmp( p1->pe.th32ProcessID, p2->pe.th32ProcessID );
}



unsigned int CountChildren( DWORD Pid, _ProcInfo *ProcessList, unsigned int NumProcesses )
{
	if ( !ProcessList || !NumProcesses )
		return 0;

	unsigned int k		= 0;

	for ( unsigned int i = 0; i < NumProcesses; i++ )
		if ( ProcessList[i].pe.th32ProcessID && ( ProcessList[i].pe.th32ParentProcessID == Pid ) )
			k++;

	return k;
}



int FixKernelPath( LPTSTR Filename )
{
	if ( !Filename )
		return -1;
	else if ( Filename[0] == '\0' )
		return 0;



	wchar_t LogicalDrives[SHORT_BUF_LEN]		= L"";
	wchar_t TargetPath[MAX_ARG_LEN]				= L"";
	wchar_t DriveLetter[4]						= L"X:";
	wchar_t temp[MAX_FILENAME_LEN]				= L"";

	LPTSTR DosDrive								= LogicalDrives;
	size_t len									= 0;



	if ( !_wcsnicmp( Filename, L"\\Device\\LanmanRedirector\\", 25 ) ) {
		wcscpy_s( temp, MAX_FILENAME_LEN, Filename + 24 );
		Filename[1] = '\0';
		wcscat_s( Filename, MAX_FILENAME_LEN, temp );
		return 0;
	}

	if ( !_wcsnicmp( Filename, L"\\Device\\Mup\\", 12 ) ) {
		wcscpy_s( temp, MAX_FILENAME_LEN, Filename + 11 );
		Filename[1] = '\0';
		wcscat_s( Filename, MAX_FILENAME_LEN, temp );
		return 0;
	}

	GetLogicalDriveStrings( SHORT_BUF_LEN - 1, LogicalDrives );
	if ( LogicalDrives[0] == '\0' )
		return 0;


	while ( DosDrive[0] ) {

		if ( iswalnum( DosDrive[0] ) && !wcscmp( DosDrive + 1, L":\\" ) ) {
			DriveLetter[0] = DosDrive[0];

			TargetPath[0] = '\0';
			QueryDosDevice( DriveLetter, TargetPath, MAX_ARG_LEN );

			if ( TargetPath[0] ) {
				len = wcslen( TargetPath );

				if ( !_wcsnicmp( Filename, TargetPath, len ) ) {
					//	wprintf( L"   * Match for %I64u characters!\n", (unsigned __int64) len );
					wcscpy_s( temp, MAX_FILENAME_LEN, Filename + len );
					wcscpy_s( Filename, MAX_FILENAME_LEN, DriveLetter );
					wcscat_s( Filename, MAX_FILENAME_LEN, temp );
					return 0;
				}
			}
		}

		DosDrive += ( wcslen( DosDrive ) + 1 );
	}

	return 0;
}



struct _ProcessInfoArgs {
	_ProcInfo *ProcessList;
	LPCTSTR SortKeys;
	LPCTSTR HighlightSpecs;
	unsigned int NumProcesses;
	unsigned int Width;
	unsigned int Flags;
	unsigned int Count;
	unsigned int NumAncestors;
	SYSTEMTIME *Now;
};



int ShowProcessInfo( _ProcInfo *Process, LPCTSTR Prefix, BOOL LastSibling, BOOL Recurse, _ProcessInfoArgs *Args )
{
	if ( !Process || !Args )
		return -1;


	wchar_t NewPrefix[SHORT_BUF_LEN]			= L"";
	wchar_t Filename[MAX_FILENAME_LEN]			= L"";
	wchar_t temp[SHORT_BUF_LEN]					= L"";
	unsigned int NumChildren					= 0;
	_ProcInfo **Children						= NULL;
	unsigned int j								= 0;
	BOOL Last									= FALSE;
	unsigned int Ascii							= ( Args->Flags & PROCTREE_FLAG_ASCII ) ? 1 : 0;
	BOOL AsciiEllipses							= ( Args->Flags & ( PROCTREE_FLAG_ASCII | PROCTREE_FLAG_ASCII_ELLIPSES ) ) ? TRUE : FALSE;
	int Columns									= Args->Width;
	SYSTEMTIME UtcTime;
	SYSTEMTIME LocalTime;
	size_t ws									= 0;
	unsigned int TimeFmt						= Args->Flags & ( PROCTREE_FLAG_INCLUDE_DATES | PROCTREE_FLAG_INCLUDE_SECONDS | PROCTREE_FLAG_INCLUDE_MS );
	int len										= 0;
	wchar_t *ch									= NULL;
	BYTE HLattr									= CL_Highlight;

	HANDLE ProcessHandle						= NULL;
	DWORD BufLen								= MAX_FILENAME_LEN;

	int rv										= 0;



	if ( Prefix ) {
		wcscpy_s( NewPrefix, SHORT_BUF_LEN, Prefix );

		if ( Args->Flags & PROCTREE_FLAG_NO_TREE_GRAPHIC )
			;
		else if ( LastSibling )
			wcscat_s( NewPrefix, SHORT_BUF_LEN, TreeGraf[2 + Ascii] );
		else
			wcscat_s( NewPrefix, SHORT_BUF_LEN, TreeGraf[0 + Ascii] );

		QPuts( NewPrefix );
		Columns -= (int) wcslen( NewPrefix );
	}
	else if ( Args->Flags & ANCESTRY_FLAG_LINE_NUMBERS ) {
		if ( Args->Flags & ANCESTRY_FLAG_REVERSE_ORDER )
			Printf( L"%3u:  ", Args->Count + 1 );
		else
			Printf( L"%3u:  ", Args->NumAncestors - Args->Count );

		Columns -= 6;
	}


	if ( Args->Flags & PROCTREE_FLAG_DECIMAL )
		wsprintf( temp, L"%u  ", Process->pe.th32ProcessID );
	else
		wsprintf( temp, L"%04X  ", Process->pe.th32ProcessID );

	QPuts( temp );
	Columns -= (int) wcslen( temp );


	if ( Args->Flags & PROCTREE_FLAG_SHOW_PARENT ) {
		if ( Args->Flags & PROCTREE_FLAG_DECIMAL )
			wsprintf( temp, L"[%u]  ", Process->pe.th32ParentProcessID );
		else
			wsprintf( temp, L"[%04X]  ", Process->pe.th32ParentProcessID );

		QPuts( temp );
		Columns -= (int) wcslen( temp );
		Args->Flags &= ~PROCTREE_FLAG_SHOW_PARENT;
	}


	if ( Args->Flags & PROCTREE_FLAG_SHOW_MEMORY_INFO ) {
		ws = Process->mi.WorkingSetSize;
		
		if ( Args->Flags & PROCTREE_FLAG_SHOW_MEMORY_BYTES )
			wsprintf( temp, L"%12I64u  ", (unsigned __int64) ws );
		else if ( ws >= ( 1024 * 1024 * 1024 ) ) {
			if ( Args->Flags & PROCTREE_FLAG_SHOW_MEMORY_FRACTIONS )
				swprintf( temp, SHORT_BUF_LEN, L"%7.2fG  ", ( (double) ws / 1024 / 1024 / 1024 ) );
			else {
				ws = ws >> 29;
				ws = ( ws >> 1 ) + ( ws & 1 );
				wsprintf( temp, L"%4I64uG  ", (unsigned __int64) ws );
			}
		}
		else if ( ws >= ( 1024 * 1024 ) ) {
			if ( Args->Flags & PROCTREE_FLAG_SHOW_MEMORY_FRACTIONS )
				swprintf( temp, SHORT_BUF_LEN, L"%7.2fM  ", ( (double) ws / 1024 / 1024 ) );
			else {
				ws = ws >> 19;
				ws = ( ws >> 1 ) + ( ws & 1 );

				wsprintf( temp, L"%4I64uM  ", (unsigned __int64) ws );
			}
		}
		else {
			if ( Args->Flags & PROCTREE_FLAG_SHOW_MEMORY_FRACTIONS )
				swprintf( temp, SHORT_BUF_LEN, L"%7.2fK  ", ( (double) ws / 1024 ) );
			else {
				ws = ws >> 9;
				ws = ( ws >> 1 ) + ( ws & 1 );
				wsprintf( temp, L"%4I64uK  ", (unsigned __int64) ws );
			}
		}

		QPuts( temp );
		Columns -= (int) wcslen( temp );
	}


#ifdef _WIN64

	if ( Args->Flags & PROCTREE_FLAG_SHOW_EXTRA_INFO ) {

		if ( Process->Elevated ) {
			if ( Ascii )
				QPuts( L"^" );
			else
				QPuts( L"\u2191" );
		}
		else
			QPuts( L" " );

		if ( Process->Wow64 ) {
			if ( Ascii )
				QPuts( L"> " );
			else
				QPuts( L"\u00bb " );
		}
		else
			QPuts( L"  " );

		Columns -= 3;
	}

#else	//	_WIN64

	if ( Args->Flags & PROCTREE_FLAG_SHOW_EXTRA_INFO ) {

		if ( Process->Elevated ) {
			if ( Ascii )
				QPuts( L"^ " );
			else
				QPuts( L"\u2191 " );
		}
		else
			QPuts( L"  " );

		Columns -= 2;
	}

#endif	//	_WIN64



	HLattr = CL_Highlight;

	if ( Args->HighlightSpecs )
		if ( MatchWildcardList( Process->pe.szExeFile, Args->HighlightSpecs, 0, 1, 1 ) )
			HLattr = CL_ReverseHighlight;

	wcscpy_s( temp, SHORT_BUF_LEN, Process->pe.szExeFile );

	if ( ( Args->Flags & PROCTREE_FLAG_SHOW_EXTENSIONS ) == 0 ) {
		ch = wcsrchr( temp, '.' );
		if ( ch )
			*ch = '\0';
	}

	len = (int) wcslen( temp );

	if ( len <= Columns ) {
		HLputs2( temp, HLattr, 0 );
		Columns -= (int) wcslen( temp );
	}
	else if ( AsciiEllipses && ( Columns > 7 ) ) {
		temp[Columns - 3] = '\0';
		wcscat_s( temp, SHORT_BUF_LEN, L"..." );
		HLputs2( temp, HLattr, 0 );
		Columns = 0;
	}
	else if ( !AsciiEllipses && ( Columns > 5 ) ) {
		temp[Columns - 1] = '\0';
		wcscat_s( temp, SHORT_BUF_LEN, L"\u2026" );
		HLputs2( temp, HLattr, 0 );
		Columns = 0;
	}
	else if ( Columns > 0 ) {
		temp[Columns] = '\0';
		HLputs2( temp, HLattr, 0 );
		Columns = 0;
	}


	if ( ( Args->Flags & PROCTREE_FLAG_SHOW_PRIORITY ) && ( Process->Priority <= 6 ) ) {
		wsprintf( temp, L"  %s", PriorityStr[Process->Priority] );
		if ( Columns >= (int) wcslen( temp ) ) {
			QPuts( temp );
			Columns -= (int) wcslen( temp );
		}
		Columns -= 4;
	}


	if ( Args->Flags & PROCTREE_FLAG_SHOW_THREAD_COUNT ) {
		wsprintf( temp, L" (%u)", Process->pe.cntThreads );
		if ( (int) wcslen( temp ) <= Columns ) {
			QPuts( temp );
			Columns -= (int) wcslen( temp );
		}
	}


	if ( Args->Flags & PROCTREE_FLAG_SHOW_TIMES ) {
		if ( Process->StartTime.dwHighDateTime | Process->StartTime.dwLowDateTime ) {

			FileTimeToSystemTime( &( Process->StartTime ), &UtcTime );
			SystemTimeToTzSpecificLocalTime( NULL, &UtcTime, &LocalTime );

			if ( ( LocalTime.wYear != Args->Now->wYear ) || ( LocalTime.wMonth != Args->Now->wMonth ) || ( LocalTime.wDay != Args->Now->wDay ) )
				TimeFmt |= PROCTREE_FLAG_INCLUDE_DATES;

			if ( TimeFmt & PROCTREE_FLAG_INCLUDE_DATES ) {
				wsprintf( temp, L"  %04u-%02u-%02u ", LocalTime.wYear, LocalTime.wMonth, LocalTime.wDay );
				len = 13;
			}
			else {
				wcscpy_s( temp, SHORT_BUF_LEN, L"  " );
				len = 2;
			}

			wsprintf( temp + len, L"%02u:%02u", LocalTime.wHour, LocalTime.wMinute );
			len += 5;

			if ( TimeFmt & PROCTREE_FLAG_INCLUDE_MS ) {
				wsprintf( temp + len, L":%02u.%03u", LocalTime.wSecond, LocalTime.wMilliseconds );
				len += 7;
			}
			else if ( TimeFmt & PROCTREE_FLAG_INCLUDE_SECONDS ) {
				wsprintf( temp + len, L":%02u", LocalTime.wSecond );
				len += 3;
			}

			if ( len <= Columns ) {
				QPuts( temp );
				Columns -= len;
			}
		}
	}


	if ( ( Args->Flags & ( PROCTREE_FLAG_SHOW_FILENAMES | PROCTREE_FLAG_NO_DISPLAY_STRINGS ) ) == PROCTREE_FLAG_SHOW_FILENAMES ) {
		ProcessHandle = OpenProcess( ( OSVerInfo.dwMajorVersion < 6 ) ? PROCESS_QUERY_INFORMATION : PROCESS_QUERY_LIMITED_INFORMATION, FALSE, Process->pe.th32ProcessID );
		if ( ProcessHandle ) {

			if ( QueryFullProcessImageNameProc ) {
				BufLen = MAX_FILENAME_LEN;
				QueryFullProcessImageNameProc( ProcessHandle, 0, Filename, &BufLen );
			}
			else {
				GetProcessImageFileName( ProcessHandle, Filename, MAX_FILENAME_LEN );
				FixKernelPath( Filename );
			}

			CloseHandle( ProcessHandle );
		}
	}


	if ( ( Args->Flags & PROCTREE_FLAG_NO_DISPLAY_STRINGS ) == 0 ) {

		HLattr = CL_Normal;
		temp[0] = '\0';

		if ( Filename[0] )
			wcscpy_s( temp, SHORT_BUF_LEN, Filename );
		else if ( Process->Title )
			wcscpy_s( temp, SHORT_BUF_LEN, Process->Title );

		if ( Args->HighlightSpecs && temp[0] )
			if ( MatchWildcardList( temp, Args->HighlightSpecs, 0, 1, 0 ) )
				HLattr = CL_ReverseHighlight;


		if ( temp[0] ) {
			len = (int) wcslen( temp );

			if ( ( len + 2 ) <= Columns ) {
				QPuts( L"  " );
				HLputs2( temp, HLattr, 0 );
				Columns -= ( len + 2 );
			}
			else if ( !AsciiEllipses && ( Columns > 7 ) ) {
				temp[Columns - 3] = 0x2026;
				temp[Columns - 2] = '\0';
				QPuts( L"  " );
				HLputs2( temp, HLattr, 0 );
				Columns = 0;
			}
			else if ( Columns > 9 ) {
				temp[Columns - 5] = '.';
				temp[Columns - 4] = '.';
				temp[Columns - 3] = '.';
				temp[Columns - 2] = '\0';
				QPuts( L"  " );
				HLputs2( temp, HLattr, 0 );
				Columns = 0;
			}
			else
				Columns = 0;
		}
	}


	if ( ( rv = CL_EndLine() ) == USER_ABORT )
		goto CleanupAndExit;


	if ( Process->Priority != UINT_MAX )
		Args->Count += 1;


	if ( Recurse )
		NumChildren = CountChildren( Process->pe.th32ProcessID, Args->ProcessList, Args->NumProcesses );

	if ( NumChildren ) {

		Children = (_ProcInfo**) malloc( NumChildren * sizeof ( _ProcInfo* ) );
		if ( !Children )
			return ERROR_NOT_ENOUGH_MEMORY;

		for ( unsigned int i = 0; i < Args->NumProcesses; i++ )
			if ( Args->ProcessList[i].pe.th32ProcessID && ( Args->ProcessList[i].pe.th32ParentProcessID == Process->pe.th32ProcessID ) )
				Children[j++] =Args->ProcessList + i;


		if ( Args->SortKeys && ( NumChildren > 1 ) )
			qsort_s( Children, NumChildren, sizeof( _ProcInfo* ), ProcTreeCompare$, (void*) Args->SortKeys );


		for ( unsigned int i = 0; i < NumChildren; i++ ) {

			Last = ( i == ( NumChildren - 1 ) );

			if ( Prefix )
				wcscpy_s( NewPrefix, SHORT_BUF_LEN, Prefix );
			else
				NewPrefix[0] = '\0';

			if ( Args->Flags & PROCTREE_FLAG_NO_TREE_GRAPHIC )
				wcscat_s( NewPrefix, SHORT_BUF_LEN, L"  " );
			else if ( LastSibling )
				wcscat_s( NewPrefix, SHORT_BUF_LEN, TreeGraf[6 + Ascii] );
			else
				wcscat_s( NewPrefix, SHORT_BUF_LEN, TreeGraf[4 + Ascii] );


			if ( rv = ShowProcessInfo( Children[i], NewPrefix, Last, TRUE, Args ) )
				goto CleanupAndExit;

		}
	}

	rv = 0;


CleanupAndExit:

	free( Children );
	return rv;
}



DWORD FindServicesRoot( _ProcInfo *ProcessList, unsigned int NumProcesses, unsigned int MinReferences )
{
	if ( !ProcessList || !NumProcesses )
		return NULL;


	const unsigned int ListSize			= 16;
	unsigned int ListCount				= 0;
	DWORD PidList[ListSize];
	unsigned int PidCount[ListSize];

	DWORD ParentPid						= 0;
	unsigned int Found					= 0;


	if ( !MinReferences )
		MinReferences = 5;


	for ( unsigned int i = 0; i < NumProcesses; i++ ) {
		if ( ProcessList[i].pe.th32ParentProcessID && !wcscmp( ProcessList[i].pe.szExeFile, L"svchost.exe" ) ) {

			ParentPid = ProcessList[i].pe.th32ParentProcessID;
			Found = UINT_MAX;

			for ( unsigned int j = 0; j < ListCount; j++ )
				if ( ParentPid == PidList[j] )
					Found = j;

			if ( Found == UINT_MAX ) {
				if ( ListCount < ListSize ) {
					PidList[ListCount]	= ParentPid;
					PidCount[ListCount]	= 1;
					ListCount++;
				}
			}
			else {
				PidCount[Found] += 1;
				if ( PidCount[Found] >= MinReferences )
					return ParentPid;
			}

		}
	}

	return NULL;
}



DWORD WildFindProcess( LPCTSTR FindName, _ProcInfo *ProcessList, unsigned int NumProcesses )
{
	if ( !FindName || !ProcessList || !NumProcesses )
		return NULL;
	else if ( FindName[0] == '\0' )
		return NULL;
	else if ( !wcscmp( FindName, L"." ) )
		return GetCurrentProcessId();


	wchar_t FindExeName[MAX_ARG_LEN]		= L"";

	DWORD MyPid				= GetCurrentProcessId();
	size_t len				= 0;


	wcscpy_s( FindExeName, MAX_ARG_LEN, FindName );

	if ( !wcschr( FindExeName, '.' ) ) {
		len = wcslen( FindExeName );

		if ( FindExeName[len - 1] != '*' )
			wcscat_s( FindExeName, MAX_ARG_LEN, L".exe" );
	}


	for ( unsigned int i = 0; i < NumProcesses; i++ )
		if ( WildcardComparison( (LPTSTR) FindExeName, ProcessList[i].pe.szExeFile, 0, 1 ) == 0 )
			return ProcessList[i].pe.th32ProcessID;

	for ( unsigned int i = 0; i < NumProcesses; i++ )
		if ( ProcessList[i].Title )
			if ( ( ProcessList[i].pe.th32ProcessID != MyPid ) && ( WildcardComparison( (LPTSTR) FindName, (LPTSTR) ProcessList[i].Title, 0, 1 ) == 0 ) )
				return ProcessList[i].pe.th32ProcessID;

	return NULL;
}



int ProcessTree( DWORD Pid, LPCTSTR FindName, LPCTSTR SortKeys, LPCTSTR HighlightSpecs, unsigned int Width, unsigned int Flags )
{
	HANDLE Snapshot				= NULL;
	DWORD SnapshotFlags			= TH32CS_SNAPPROCESS;


	_ProcInfo *ProcessList		= NULL;
	_ProcInfo *RootProcess		= NULL;
	_ProcInfo DummyRoot;

	_PidSvcName *PidSvcNames	= NULL;
	_PidWinTitle *PidWinTitles	= NULL;

	HANDLE ProcessHandle		= NULL;
	FILETIME ExitTime, KernelTime, UserTime;
	SYSTEMTIME Now;

	_ProcessInfoArgs Args;

	unsigned int NumProcesses	= 0;

	int rv						= 0;



	GetLocalTime( &Now );

	Snapshot = CreateToolhelp32Snapshot( SnapshotFlags, 0 );
	if ( !Snapshot || ( Snapshot == INVALID_HANDLE_VALUE ) )
		goto ErrorExit;


	NumProcesses = CountProcessEntries( Snapshot );
	if ( !NumProcesses ) {
		Printf( L"No processes found!\r\n" );
		rv = 0;
		goto CleanupAndExit;
	}


	ProcessList = (_ProcInfo*) malloc( NumProcesses * sizeof( _ProcInfo ) );
	if ( !ProcessList ) {
		rv = ERROR_NOT_ENOUGH_MEMORY;
		goto CleanupAndExit;
	}

	memset( ProcessList, 0x00, NumProcesses * sizeof( _ProcInfo ) );
	for ( unsigned int i = 0; i < NumProcesses; i++ ) {
		ProcessList[i].pe.dwSize	= sizeof( PROCESSENTRY32 );
		ProcessList[i].mi.cb		= sizeof( PROCESS_MEMORY_COUNTERS );
	}


	if ( !Process32First( Snapshot, &( ProcessList[0].pe ) ) )
		goto ErrorExit;

	for ( unsigned int i = 1; i < NumProcesses; i++ )
		if ( !Process32Next( Snapshot, &( ProcessList[i].pe ) ) )
			goto ErrorExit;

	for ( unsigned int i = 0; i < NumProcesses; i++ )
		if ( ProcessList[i].pe.th32ProcessID && ProcessList[i].pe.th32ParentProcessID )
			if ( ProcTree_ProcessIsNewer( ProcessList[i].pe.th32ParentProcessID, ProcessList[i].pe.th32ProcessID ) )
				ProcessList[i].pe.th32ParentProcessID = 0;

	if ( Flags & PROCTREE_FLAG_GET_EXTRA_INFO ) {
		for ( unsigned int i = 0; i < NumProcesses; i++ ) {
			ProcessHandle = OpenProcess( ( OSVerInfo.dwMajorVersion < 6 ) ? PROCESS_QUERY_INFORMATION : PROCESS_QUERY_LIMITED_INFORMATION, FALSE, ProcessList[i].pe.th32ProcessID );
			if ( ProcessHandle ) {

#ifdef _WIN64

				if ( IsWow64Process( ProcessHandle, &( ProcessList[i].Wow64 ) ) == 0 )
					ProcessList[i].Wow64 = FALSE;

#endif	//	_WIN64

				if ( ProcessElevated( ProcessHandle ) )
					ProcessList[i].Elevated = TRUE;

				GetProcessTimes( ProcessHandle, &( ProcessList[i].StartTime ), &ExitTime, &KernelTime, &UserTime );

				ProcessList[i].Priority = PriorityLevel( GetPriorityClass( ProcessHandle ) );

				GetProcessMemoryInfo( ProcessHandle, &( ProcessList[i].mi ), sizeof( PROCESS_MEMORY_COUNTERS ) );

				CloseHandle( ProcessHandle );
				ProcessHandle = NULL;
			}
		}
	}


	if ( ( Flags & PROCTREE_FLAG_NO_SERVICE_NAMES ) == 0 )
		CollectPidServiceNames( &PidSvcNames );

	if ( ( Flags & PROCTREE_FLAG_NO_WINDOW_TITLES ) == 0 )
		CollectPidWinTitles( &PidWinTitles );

	for ( unsigned int i = 0; i < NumProcesses; i++ ) {
		if ( PidSvcNames )
			ProcessList[i].Title	= SvcNamesForPid( ProcessList[i].pe.th32ProcessID, PidSvcNames );

		if ( PidWinTitles && ( ProcessList[i].Title == NULL ) )
			ProcessList[i].Title	= PidWinTitle( ProcessList[i].pe.th32ProcessID, PidWinTitles );
	}


	if ( Flags & PROCTREE_FLAG_FIND_SERVICES ) {
		Pid = FindServicesRoot( ProcessList, NumProcesses, 5 );
		if ( !Pid ) {
			SetLastError( PROCTREE_ERROR_SERVICES_NOT_FOUND );
			goto ErrorExit;
		}
	}
	else if ( FindName ) {
		Pid = WildFindProcess( FindName, ProcessList, NumProcesses );
		if ( !Pid ) {
			SetLastError( PROCTREE_ERROR_NO_MATCH );
			goto ErrorExit;
		}
	}


	RootProcess = FindProcessEntry( Pid, ProcessList, NumProcesses );

	if ( !RootProcess ) {
		if ( CountChildren( Pid, ProcessList, NumProcesses ) > 0 ) {
			memset( &DummyRoot, 0x00, sizeof( _ProcInfo ) );
			DummyRoot.pe.dwSize			= sizeof( PROCESSENTRY32 );
			DummyRoot.pe.th32ProcessID	= Pid;
			DummyRoot.Priority			= UINT_MAX;
			wcscpy_s( DummyRoot.pe.szExeFile, MAX_PATH, L"* PROCESS HAS ENDED *" );
			RootProcess = &DummyRoot;
			Flags &= ~PROCTREE_FLAG_SHOW_PARENT;
		}
	}

	if ( !RootProcess ) {
		SetLastError( PROCTREE_ERROR_PID_NOT_FOUND );
		goto ErrorExit;
	}

	if ( HighlightSpecs )
		if ( HighlightSpecs[0] == '\0' )
			HighlightSpecs = NULL;


	Args.ProcessList	= ProcessList;
	Args.SortKeys		= SortKeys;
	Args.HighlightSpecs	= HighlightSpecs;
	Args.NumProcesses	= NumProcesses;
	Args.Width			= Width;
	Args.Flags			= Flags;
	Args.Count			= 0;
	Args.NumAncestors	= 0;
	Args.Now			= &Now;


	if ( ( Args.Flags & PROCTREE_FLAG_NO_TREE_GRAPHIC ) == 0 )
		CrLf();

	rv = ShowProcessInfo( RootProcess, NULL, TRUE, TRUE, &Args );

	if ( rv == USER_ABORT )
		goto CleanupAndExit;
	else if ( rv )
		goto ErrorExit;


	if ( Args.Flags & PROCTREE_FLAG_SHOW_SUMMARY ) {
		if ( ( rv = HLprintf( L"\nFound %u process%s.\n", Args.Count, ( Args.Count == 1 ) ? L"" : L"es" ) ) == USER_ABORT )
			goto CleanupAndExit;
	}

	rv = 0;
	goto CleanupAndExit;


ErrorExit:

	rv = GetLastError();
	if ( !rv )
		rv = -1;

CleanupAndExit :

	free( ProcessList );
	CloseHandle( Snapshot );
	FreeLinkedList( (void**) &PidSvcNames );
	FreeLinkedList( (void**) &PidWinTitles );
	return rv;
}



BOOL CALLBACK FindDesktopWindow$( HWND hWnd, LPARAM lParam )
{
	wchar_t ClassName[SHORT_BUF_LEN]		= L"";
	HWND *DesktopWindow						= (HWND*) lParam;

	RealGetWindowClass( hWnd, ClassName, SHORT_BUF_LEN );

	if ( !wcscmp( ClassName, L"Progman" ) ) {
		*DesktopWindow = hWnd;
		return FALSE;
	}

	return TRUE;
}



DWORD GetDesktopPid()
{
	HWND DesktopWindow		= NULL;
	DWORD DesktopPid		= NULL;


	EnumWindows( FindDesktopWindow$, (LPARAM) &DesktopWindow );
	if ( !DesktopWindow )
		return NULL;

	GetWindowThreadProcessId( DesktopWindow, &DesktopPid );
	return DesktopPid;
}



HWND GetTakeCmdWindow()
{
	wchar_t temp[SHORT_BUF_LEN]		= L"";

	HWND rv							= NULL;


	TakeCommandIPC( L"HWND", temp );

	if ( iswdigit( temp[0] ) )
		rv = (HWND) wcstoul( temp, NULL, 0 );

	return rv;
}



DWORD GetTakeCmdPid()
{
	HWND TakeCmdWindow		= NULL;
	DWORD TakeCmdPid		= NULL;


	TakeCmdWindow = GetTakeCmdWindow();
	if ( !TakeCmdWindow )
		return NULL;

	GetWindowThreadProcessId( TakeCmdWindow, &TakeCmdPid );
	return TakeCmdPid;
}



int InvertCase( LPTSTR String )
{
	if ( !String )
		return -1;


	int k			= 0;


	for ( wchar_t *ch = String; *ch; ch++ ) {
		if ( iswupper( *ch ) )
			*ch = towlower( *ch ), k++;
		else if ( iswlower( *ch ) )
			*ch = towupper( *ch ), k++;
	}

	return k;
}



int ParseOOption( LPCTSTR inString, LPTSTR SortKeys, unsigned int SortKeyLen, BOOL *Unsorted )
{
	if ( !inString || !SortKeys || SortKeyLen < 5 )
		return -1;


	const wchar_t *ch	= inString;
	unsigned int k		= 0;
	BOOL rev			= FALSE;
	BOOL revall			= FALSE;


	SortKeys[k]		= '\0';

	if ( Unsorted )
		*Unsorted = FALSE;


	if ( IsColonOrEquals( *ch ) )
		ch++;

	
	if ( towupper( *ch ) == 'U' ) {
		SortKeys[k] = '\0';
		
		if ( Unsorted )
			*Unsorted = TRUE;
	}
	else {
		while ( *ch ) {

			switch ( towupper( *ch ) ) {

			case 'N' :
			case 'P' :
			case 'T' :
			case 'K' :
			case 'E' :
			case 'W' :
			case 'D' :
			case 'C' :
			case 'M' :

				if ( k < ( SortKeyLen - 1 ) ) {
					SortKeys[k++] = rev ? towupper( *ch ) : towlower( *ch );
					SortKeys[k] = '\0';
				}

				rev = FALSE;
				break;

			case 'R' :

				revall = TRUE;
				rev = FALSE;
				break;

			case '-' :

				rev = !rev;
				break;

			default :

				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Unknown sort flag: /O:%c  (/O? for help)\r\n", towupper( *ch ) );
				return 1;
			}

			ch++;
		}

		if ( revall )
			InvertCase( SortKeys );
	}

	return 0;
}



int SortingRequiresExtraInfo( LPCTSTR SortKeys )
{
	if ( !SortKeys )
		return -1;


	const wchar_t *ch		= SortKeys;
	wchar_t uc				= '\0';
	int rv					= 0;


	while ( *ch ) {
		uc = towupper( *ch );

		if ( ( uc == 'T' ) || ( uc == 'E' ) || ( uc == 'W' ) || ( uc = 'C' ) || ( uc == 'M' ) )
			rv++;

		ch++;
	}

	return rv;
}



int ShowCmdHelp( LPCTSTR HelpText, BOOL ShowVersion )
{
	if ( !HelpText )
		return 1;

	QPuts( (LPTSTR) HelpText );

	if ( ShowVersion )
		Printf( L"\r\n\t\t\t\t%s  %u.%u.%u.%u (%s)\r\n", PLUGIN_NAME, VER_MAJOR, VER_MINOR, VER_BUILD, VER_PATCH, PLATFORM );

	return 0;
}



DLLExports int WINAPI proctree( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t ErrorText[SHORT_BUF_LEN]			= L"";
	wchar_t SortKeys[PROCTREE_SORTKEY_LEN]		= L"";
	wchar_t HighlightSpecs[MAX_ARG_LEN]			= L"";

	LPCTSTR FindName							= NULL;

	DWORD Pid							= 0;
	unsigned int Flags					= 0;
	BOOL GotPid							= FALSE;
	BOOL NoHighlight					= FALSE;
	BOOL UseShellPid					= FALSE;
	BOOL UseTakeCmdPid					= FALSE;
	BOOL Unsorted						= FALSE;
	unsigned int Pause					= 0;
	unsigned int Width					= 0;
	wchar_t *sep1						= NULL;

	int rv					= 0;



	CL_NoPausing();

	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE );


	for ( int i = 0; i < numargs; i++ ) {

		if ( !wcsncmp( arg[i], L"/?", 2 ) )
			return ShowCmdHelp( ProcTreeHelpText, ( arg[i][2] == '?' ) );
		else if ( !_wcsicmp( arg[i], L"/A" ) )
			Flags |= PROCTREE_FLAG_ASCII;
		else if ( !_wcsicmp( arg[i], L"/C" ) )
			Flags |= ( PROCTREE_FLAG_GET_EXTRA_INFO | PROCTREE_FLAG_SHOW_PRIORITY );
		else if ( !_wcsicmp( arg[i], L"/D" ) )
			Flags |= PROCTREE_FLAG_DECIMAL;
		else if ( !_wcsicmp( arg[i], L"/E" ) )
			UseShellPid = TRUE;
		else if ( !_wcsicmp( arg[i], L"/F" ) )
			Flags |= PROCTREE_FLAG_SHOW_FILENAMES;
		else if ( !_wcsnicmp( arg[i], L"/H", 2 ) ) {
			int j = 2;
			if ( IsColonOrEquals( arg[i][j] ) )
				j++;
			else if ( ( towupper( arg[i][j] ) == 'L' ) && IsColonOrEquals( arg[i][j + 1] ) )
				j += 2;

			if ( arg[i][j] ) {
				if ( HighlightSpecs[0] )
					wcscat_s( HighlightSpecs, MAX_ARG_LEN, L"|" );

				wcscat_s( HighlightSpecs, MAX_ARG_LEN, arg[i] + j );
			}
		}
		else if ( !_wcsicmp( arg[i], L"/I" ) )
			Flags |= ( PROCTREE_FLAG_GET_EXTRA_INFO | PROCTREE_FLAG_SHOW_EXTRA_INFO );
		else if ( !_wcsicmp( arg[i], L"/K" ) )
			Flags |= PROCTREE_FLAG_SHOW_THREAD_COUNT;
		else if ( !_wcsicmp( arg[i], L"/M" ) )
			Flags = ( Flags & ~( PROCTREE_FLAG_SHOW_MEMORY_BYTES | PROCTREE_FLAG_SHOW_MEMORY_FRACTIONS ) ) | PROCTREE_FLAG_GET_EXTRA_INFO | PROCTREE_FLAG_SHOW_MEMORY_INFO;
		else if ( !_wcsicmp( arg[i], L"/MF" ) )
			Flags = ( Flags & !PROCTREE_FLAG_SHOW_MEMORY_BYTES ) | PROCTREE_FLAG_GET_EXTRA_INFO | PROCTREE_FLAG_SHOW_MEMORY_INFO | PROCTREE_FLAG_SHOW_MEMORY_FRACTIONS;
		else if ( !_wcsicmp( arg[i], L"/MM" ) )
			Flags = ( Flags & !PROCTREE_FLAG_SHOW_MEMORY_FRACTIONS ) | PROCTREE_FLAG_GET_EXTRA_INFO | PROCTREE_FLAG_SHOW_MEMORY_INFO | PROCTREE_FLAG_SHOW_MEMORY_BYTES;
		else if ( !_wcsnicmp( arg[i], L"/N", 2 ) ) {
			wchar_t *ch		= arg[i] + 2;
			if ( IsColonOrEquals( *ch ) )
				ch++;

			while ( *ch ) {

				switch ( towupper( *ch ) ) {

				case 'C' :
					NoHighlight = TRUE;
					break;

				case 'D' :
					Flags |= PROCTREE_FLAG_NO_DISPLAY_STRINGS;
					break;

				case 'E' :
					Flags |= PROCTREE_FLAG_ASCII_ELLIPSES;
					break;

				case 'S' :
					Flags |= PROCTREE_FLAG_NO_SERVICE_NAMES;
					break;

				case 'T' :
					Flags |= PROCTREE_FLAG_NO_TREE_GRAPHIC;
					break;

				case 'W' :
					Flags |= PROCTREE_FLAG_NO_WINDOW_TITLES;
					break;

				case 'X' :
					Flags |= PROCTREE_FLAG_SHOW_EXTENSIONS;
					break;

				case '?' :
					QPuts( (LPTSTR) ProcTreeOptNHelpText );
					return 0;

				default :
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Unknown suboption: /N:%c  (/N? for help)\r\n", towupper( *ch ) );
					return 1;
				}

				ch++;
			}
		}
		else if ( !_wcsnicmp( arg[i], L"/O", 2 ) ) {
			if ( wcschr( arg[i], '?' ) ) {
				QPuts( (LPTSTR) ProcTreeOptOHelpText );
				return 0;
			}

			if ( ParseOOption( arg[i] + 2, SortKeys, PROCTREE_SORTKEY_LEN, &Unsorted ) )
				return 1;
		}
		else if ( !_wcsnicmp( arg[i], L"/P", 2 ) ) {
			if ( CL_ParsePOption( arg[i], &Pause ) )
				return 1;
		}
		else if ( !_wcsicmp( arg[i], L"/SV" ) || !_wcsicmp( arg[i], L"/SVC" ) )
			Flags |= PROCTREE_FLAG_FIND_SERVICES;
		else if ( !_wcsicmp( arg[i], L"/TC" ) || !_wcsicmp( arg[i], L"/TCMD" ) )
			UseTakeCmdPid = TRUE;
		else if ( !_wcsnicmp( arg[i], L"/T", 2 ) ) {
			Flags |= ( PROCTREE_FLAG_GET_EXTRA_INFO | PROCTREE_FLAG_SHOW_TIMES );

			int j = 2;
			if ( IsColonOrEquals( arg[i][j] ) )
				j++;

			while ( arg[i][j] ) {

				switch ( towupper( arg[i][j] ) ) {

				case 'D' :
					Flags |= PROCTREE_FLAG_INCLUDE_DATES;
					break;

				case 'S' :
					Flags |= PROCTREE_FLAG_INCLUDE_SECONDS;
					break;

				case 'M' :
					Flags |= ( PROCTREE_FLAG_INCLUDE_SECONDS | PROCTREE_FLAG_INCLUDE_MS );
					break;

				case 'F' :
					Flags |= ( PROCTREE_FLAG_INCLUDE_DATES | PROCTREE_FLAG_INCLUDE_SECONDS | PROCTREE_FLAG_INCLUDE_MS );
					break;

				case '?' :
					QPuts( (LPTSTR) ProcTreeOptTHelpText );
					return 0;

				default :
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Unknown suboption: /T:%c  (/T? for help)\r\n", towupper( arg[i][j] ) );
					return 1;
				}

				j++;
			}
		}
		else if ( !_wcsicmp( arg[i], L"/U" ) )
			Flags |= PROCTREE_FLAG_SHOW_SUMMARY;
		else if ( !_wcsicmp( arg[i], L"/W" ) )
			Width = 0;
		else if ( !_wcsnicmp( arg[i], L"/W", 2 ) ) {
			int j = 2;
			if ( IsColonOrEquals( arg[i][j] ) )
				j++;

			if ( ParseInt( arg[i] + j, &Width, NULL ) ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad width: \"%s\"\r\n", arg[i] + j );
				return 1;
			}

			if ( Width && ( ( Width < PROCTREE_WIDTH_MIN ) || ( Width > PROCTREE_WIDTH_MAX ) ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Width must be between %u and %u: %u\r\n", PROCTREE_WIDTH_MIN, PROCTREE_WIDTH_MAX, Width );
				return 1;
			}
		}
		else if ( !_wcsicmp( arg[i], L"/X" ) )
			Flags &= ~PROCTREE_FLAG_DECIMAL;
		else if ( !_wcsicmp( arg[i], L"/Z" ) )
			Flags |= PROCTREE_FLAG_SHOW_PARENT;
		else if ( arg[i][0] == '/' ) {
			_wcsupr_s( arg[i], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
			return 1;
		}
		else if ( !GotPid && !FindName ) {
			sep1 = NULL;

			if ( ParseInt( arg[i], (unsigned int*) &Pid, &sep1 ) )
				FindName = arg[i];
			else
				GotPid = TRUE;

			if ( sep1 )
				if ( *sep1 )
					FindName = arg[i], Pid = 0;

			if ( FindName )
				GotPid = TRUE;
		}
		else {
			_wcsupr_s( arg[i], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
			return 1;
		}
	}


	if ( !GotPid && UseShellPid ) {
		if ( Pid = GetDesktopPid() )
			GotPid = TRUE;
		else {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Explorer process not found.\r\n" );
			return 2;
		}
	}


	if ( !GotPid && UseTakeCmdPid )
		if ( Pid = GetTakeCmdPid() )
			GotPid = TRUE;


	if ( !GotPid && ( Flags & PROCTREE_FLAG_FIND_SERVICES ) )
		Pid = -1, GotPid = TRUE;


	if ( !GotPid )
		Pid = GetCurrentProcessId();


	if ( SortingRequiresExtraInfo( SortKeys ) )
		Flags |= PROCTREE_FLAG_GET_EXTRA_INFO;


	if ( Pause )
		CL_SetPauseLines( Pause );

	CL_PickHighlightColor( NoHighlight );

	if ( NoHighlight )
		HighlightSpecs[0] = '\0';


	if ( !Width ) {
		if ( CL_ConsoleHandle( STD_OUTPUT_HANDLE ) ) {
			Width = CL_GetConsoleWidth( PROCTREE_WIDTH_DEFAULT + 1 ) - 1;

			if ( Width < PROCTREE_WIDTH_MIN )
				Width = PROCTREE_WIDTH_MIN;
			else if ( Width > PROCTREE_WIDTH_MAX )
				Width = PROCTREE_WIDTH_MAX;
		}
		else
			Width = PROCTREE_WIDTH_DEFAULT;
	}


	if ( rv = ProcessTree( Pid, FindName, Unsorted ? NULL : SortKeys, HighlightSpecs[0] ? HighlightSpecs : NULL, Width, Flags ) ) {

		if ( rv == USER_ABORT )
			return 3;
		else if ( rv == PROCTREE_ERROR_PID_NOT_FOUND )
			wsprintf( ErrorText, L"Process ID not found: 0x%04X (%u)", Pid, Pid );
		else if ( rv == PROCTREE_ERROR_NO_MATCH )
			wcscpy_s( ErrorText, SHORT_BUF_LEN, L"No process matching the given wildcards was found." );
		else if ( rv == PROCTREE_ERROR_SERVICES_NOT_FOUND )
			wcscpy_s( ErrorText, SHORT_BUF_LEN, L"Services root not found." );
		else
			GetErrorText( rv, ErrorText );

		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"%s\r\n", ErrorText );
		return 2;
	}


	return rv;
}



int ShowAncestry( DWORD Pid, LPCTSTR FindName, LPCTSTR HighlightSpecs, unsigned int Width, unsigned int Flags )
{
	HANDLE Snapshot				= NULL;

	_ProcInfo *ProcessList		= NULL;
	_ProcInfo *ThisProcess		= NULL;
	
	_PidSvcName *PidSvcNames	= NULL;
	_PidWinTitle *PidWinTitles	= NULL;

	HANDLE ProcessHandle		= NULL;
	FILETIME ExitTime, KernelTime, UserTime;
	SYSTEMTIME Now;

	DWORD *Ancestors			= NULL;
	unsigned int NumAncestors	= 0;

	_ProcessInfoArgs Args;
	unsigned int Start			= 0;
	int Dir						= +1;

	unsigned int NumProcesses	= 0;

	int rv						= 0;



	GetLocalTime( &Now );

	Snapshot = CreateToolhelp32Snapshot( TH32CS_SNAPPROCESS, 0 );
	if ( !Snapshot || ( Snapshot == INVALID_HANDLE_VALUE ) )
		goto ErrorExit;


	NumProcesses = CountProcessEntries( Snapshot );
	if ( !NumProcesses ) {
		Printf( L"No processes found!\r\n" );
		rv = 0;
		goto CleanupAndExit;
	}


	ProcessList = (_ProcInfo*) malloc( NumProcesses * sizeof( _ProcInfo ) );
	if ( !ProcessList ) {
		rv = ERROR_NOT_ENOUGH_MEMORY;
		goto CleanupAndExit;
	}

	memset( ProcessList, 0x00, NumProcesses * sizeof( _ProcInfo ) );
	for ( unsigned int i = 0; i < NumProcesses; i++ ) {
		ProcessList[i].pe.dwSize	= sizeof( PROCESSENTRY32 );
		ProcessList[i].mi.cb		= sizeof( PROCESS_MEMORY_COUNTERS );
	}


	if ( !Process32First( Snapshot, &( ProcessList[0].pe ) ) )
		goto ErrorExit;

	for ( unsigned int i = 1; i < NumProcesses; i++ )
		if ( !Process32Next( Snapshot, &( ProcessList[i].pe ) ) )
			goto ErrorExit;

	for ( unsigned int i = 0; i < NumProcesses; i++ )
		if ( ProcessList[i].pe.th32ProcessID && ProcessList[i].pe.th32ParentProcessID )
			if ( ProcTree_ProcessIsNewer( ProcessList[i].pe.th32ParentProcessID, ProcessList[i].pe.th32ProcessID ) )
				ProcessList[i].pe.th32ParentProcessID = 0;

	if ( Flags & PROCTREE_FLAG_GET_EXTRA_INFO ) {
		for ( unsigned int i = 0; i < NumProcesses; i++ ) {
			ProcessHandle = OpenProcess( ( OSVerInfo.dwMajorVersion < 6 ) ? PROCESS_QUERY_INFORMATION : PROCESS_QUERY_LIMITED_INFORMATION, FALSE, ProcessList[i].pe.th32ProcessID );
			if ( ProcessHandle ) {

#ifdef _WIN64

				if ( IsWow64Process( ProcessHandle, &( ProcessList[i].Wow64 ) ) == 0 )
					ProcessList[i].Wow64 = FALSE;

#endif	//	_WIN64

				if ( ProcessElevated( ProcessHandle ) )
					ProcessList[i].Elevated = TRUE;

				GetProcessTimes( ProcessHandle, &( ProcessList[i].StartTime ), &ExitTime, &KernelTime, &UserTime );

				ProcessList[i].Priority = PriorityLevel( GetPriorityClass( ProcessHandle ) );

				GetProcessMemoryInfo( ProcessHandle, &( ProcessList[i].mi ), sizeof( PROCESS_MEMORY_COUNTERS ) );

				CloseHandle( ProcessHandle );
				ProcessHandle = NULL;
			}
		}
	}


	if ( ( Flags & PROCTREE_FLAG_NO_SERVICE_NAMES ) == 0 )
		CollectPidServiceNames( &PidSvcNames );

	if ( ( Flags & PROCTREE_FLAG_NO_WINDOW_TITLES ) == 0 )
		CollectPidWinTitles( &PidWinTitles );

	for ( unsigned int i = 0; i < NumProcesses; i++ ) {
		if ( PidSvcNames )
			ProcessList[i].Title	= SvcNamesForPid( ProcessList[i].pe.th32ProcessID, PidSvcNames );

		if ( PidWinTitles && ( ProcessList[i].Title == NULL ) )
			ProcessList[i].Title	= PidWinTitle( ProcessList[i].pe.th32ProcessID, PidWinTitles );
	}


	if ( FindName ) {
		Pid = WildFindProcess( FindName, ProcessList, NumProcesses );
		if ( !Pid ) {
			SetLastError( PROCTREE_ERROR_NO_MATCH );
			goto ErrorExit;
		}
	}


	if ( !GetAncestors( Pid, &Ancestors, &NumAncestors ) ) {
		SetLastError( PROCTREE_ERROR_PID_NOT_FOUND );
		goto ErrorExit;
	}


	if ( HighlightSpecs )
		if ( HighlightSpecs[0] == '\0' )
			HighlightSpecs = NULL;


	Args.ProcessList	= ProcessList;
	Args.SortKeys		= NULL;
	Args.HighlightSpecs	= HighlightSpecs;
	Args.NumProcesses	= NumProcesses;
	Args.Width			= Width;
	Args.Flags			= Flags;
	Args.Count			= 0;
	Args.NumAncestors	= NumAncestors;
	Args.Now			= &Now;


	if ( Flags & ANCESTRY_FLAG_REVERSE_ORDER )
		Start = 0, Dir = +1;
	else
		Start = NumAncestors - 1, Dir = -1;


	for ( unsigned int i = Start; i < NumAncestors; i += Dir ) {
		if ( ThisProcess = FindProcessEntry( Ancestors[i], ProcessList, NumProcesses ) ) {
			rv = ShowProcessInfo( ThisProcess, NULL, FALSE, FALSE, &Args );

			if ( rv == USER_ABORT )
				goto CleanupAndExit;
			else if ( rv )
				goto ErrorExit;
		}
	}

	if ( Args.Flags & PROCTREE_FLAG_SHOW_SUMMARY ) {
		if ( ( rv = HLprintf( L"\nFound %u process%s.\n", Args.Count, ( Args.Count == 1 ) ? L"" : L"es" ) ) == USER_ABORT )
			goto CleanupAndExit;
	}

	rv = 0;
	goto CleanupAndExit;


ErrorExit:

	rv = GetLastError();
	if ( !rv )
		rv = -1;

CleanupAndExit:

	free( Ancestors );
	free( ProcessList );
	CloseHandle( Snapshot );
	FreeLinkedList( (void**) &PidSvcNames );
	FreeLinkedList( (void**) &PidWinTitles );
	return rv;
}



DLLExports int ancestry( LPTSTR inString )
{
	if ( !inString )
		return 1;


	wchar_t ErrorText[SHORT_BUF_LEN]		= L"";
	wchar_t HighlightSpecs[MAX_ARG_LEN]		= L"";

	LPCTSTR FindName						= NULL;
	LPTSTR Title							= NULL;

	unsigned int Flags			= 0;
	DWORD Pid					= 0;
	BOOL GotPid					= FALSE;
	BOOL NoHighlight			= FALSE;
	unsigned int Pause			= 0;
	unsigned int Width			= 0;

	wchar_t *sep1				= NULL;

	int rv						= 0;



	CL_NoPausing();

	ParseArgs( inString, PARSE_BREAK_SPACES | PARSE_SLASHES_KLUDGE | PARSE_FORCE_UPPERCASE );


	for ( int i = 0; i < numargs; i++ ) {
		if ( !wcsncmp( arg[i], L"/?", 2 ) )
			return ShowCmdHelp( AncestryHelpText, ( arg[i][2] == '?' ) );
		else if ( !wcscmp( arg[i], L"/C" ) )
			Flags |= ( PROCTREE_FLAG_GET_EXTRA_INFO | PROCTREE_FLAG_SHOW_PRIORITY );
		else if ( !wcscmp( arg[i], L"/D" ) )
			Flags |= PROCTREE_FLAG_DECIMAL;
		else if ( !wcscmp( arg[i], L"/F" ) )
			Flags |= PROCTREE_FLAG_SHOW_FILENAMES;
		else if ( !wcsncmp( arg[i], L"/H", 2 ) ) {
			int j = 2;
			if ( IsColonOrEquals( arg[i][j] ) )
				j++;
			else if ( ( arg[i][j] == 'L' ) && IsColonOrEquals( arg[i][j + 1] ) )
				j += 2;

			if ( arg[i][j] ) {
				if ( HighlightSpecs[0] )
					wcscat_s( HighlightSpecs, MAX_ARG_LEN, L"|" );

				wcscat_s( HighlightSpecs, MAX_ARG_LEN, arg[i] + j );
			}
		}
		else if ( !wcscmp( arg[i], L"/I" ) )
			Flags |= ( PROCTREE_FLAG_GET_EXTRA_INFO | PROCTREE_FLAG_SHOW_EXTRA_INFO );
		else if ( !wcscmp( arg[i], L"/K" ) )
			Flags |= PROCTREE_FLAG_SHOW_THREAD_COUNT;
		else if ( !wcscmp( arg[i], L"/L" ) )
			Flags |= ANCESTRY_FLAG_LINE_NUMBERS;
		else if ( !wcscmp( arg[i], L"/M" ) )
			Flags = ( Flags & ~( PROCTREE_FLAG_SHOW_MEMORY_BYTES | PROCTREE_FLAG_SHOW_MEMORY_FRACTIONS ) ) | PROCTREE_FLAG_GET_EXTRA_INFO | PROCTREE_FLAG_SHOW_MEMORY_INFO;
		else if ( !wcscmp( arg[i], L"/MF" ) )
			Flags = ( Flags & !PROCTREE_FLAG_SHOW_MEMORY_BYTES ) | PROCTREE_FLAG_GET_EXTRA_INFO | PROCTREE_FLAG_SHOW_MEMORY_INFO | PROCTREE_FLAG_SHOW_MEMORY_FRACTIONS;
		else if ( !wcscmp( arg[i], L"/MM" ) )
			Flags = ( Flags & !PROCTREE_FLAG_SHOW_MEMORY_FRACTIONS ) | PROCTREE_FLAG_GET_EXTRA_INFO | PROCTREE_FLAG_SHOW_MEMORY_INFO | PROCTREE_FLAG_SHOW_MEMORY_BYTES;
		else if ( !wcsncmp( arg[i], L"/N", 2 ) ) {
			wchar_t *ch		= arg[i] + 2;
			if ( IsColonOrEquals( *ch ) )
				ch++;

			while ( *ch ) {

				switch ( *ch ) {

				case 'C' :
					NoHighlight = TRUE;
					break;

				case 'D' :
					Flags |= PROCTREE_FLAG_NO_DISPLAY_STRINGS;
					break;

				case 'E' :
					Flags |= PROCTREE_FLAG_ASCII_ELLIPSES;
					break;

				case 'S' :
					Flags |= PROCTREE_FLAG_NO_SERVICE_NAMES;
					break;

				case 'W' :
					Flags |= PROCTREE_FLAG_NO_WINDOW_TITLES;
					break;

				case 'X' :
					Flags |= PROCTREE_FLAG_SHOW_EXTENSIONS;
					break;

				case '?' :
					QPuts( (LPTSTR) AncestryOptNHelpText );
					return 0;

				default :
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Unknown suboptions: /N:%c  (/N? for help)\r\n", *ch );
					return 1;
				}

				ch++;
			}
		}
		else if ( !wcsncmp( arg[i], L"/P", 2 ) ) {
			if ( CL_ParsePOption( arg[i], &Pause ) )
				return 1;
		}
		else if ( !wcscmp( arg[i], L"/R" ) )
			Flags |= ANCESTRY_FLAG_REVERSE_ORDER;
		else if ( !wcsncmp( arg[i], L"/T", 2 ) ) {
			Flags |= ( PROCTREE_FLAG_GET_EXTRA_INFO | PROCTREE_FLAG_SHOW_TIMES );

			int j = 2;
			if ( IsColonOrEquals( arg[i][j] ) )
				j++;

			while ( arg[i][j] ) {

				switch ( towupper( arg[i][j] ) ) {

				case 'D' :
					Flags |= PROCTREE_FLAG_INCLUDE_DATES;
					break;

				case 'S' :
					Flags |= PROCTREE_FLAG_INCLUDE_SECONDS;
					break;

				case 'M' :
					Flags |= ( PROCTREE_FLAG_INCLUDE_SECONDS | PROCTREE_FLAG_INCLUDE_MS );
					break;

				case 'F' :
					Flags |= ( PROCTREE_FLAG_INCLUDE_DATES | PROCTREE_FLAG_INCLUDE_SECONDS | PROCTREE_FLAG_INCLUDE_MS );
					break;

				case '?' :
					QPuts( (LPTSTR) AncestryOptTHelpText );
					return 0;

				default :
					DisplayErrorHeader();
					Qprintf( ERRHANDLE, L"Unknown suboption: /T:%c  (/T? for help)\r\n", towupper( arg[i][j] ) );
					return 1;
				}

				j++;
			}
		}
		else if ( !wcscmp( arg[i], L"/U" ) )
			Flags |= PROCTREE_FLAG_SHOW_SUMMARY;
		else if ( !wcscmp( arg[i], L"/W" ) )
			Width = 0;
		else if ( !wcsncmp( arg[i], L"/W", 2 ) ) {
			int j = 2;
			if ( IsColonOrEquals( arg[i][j] ) )
				j++;

			if ( ParseInt( arg[i] + j, &Width, NULL ) ) {
				_wcsupr_s( arg[i], MAX_ARG_LEN );
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Bad width: \"%s\"\r\n", arg[i] + j );
				return 1;
			}

			if ( Width && ( ( Width < PROCTREE_WIDTH_MIN ) || ( Width > PROCTREE_WIDTH_MAX ) ) ) {
				DisplayErrorHeader();
				Qprintf( ERRHANDLE, L"Width must be between %u and %u: %u\r\n", PROCTREE_WIDTH_MIN, PROCTREE_WIDTH_MAX, Width );
				return 1;
			}
		}
		else if ( !wcscmp( arg[i], L"/X" ) )
			Flags &= ~PROCTREE_FLAG_DECIMAL;
 		else if ( arg[i][0] == '/' ) {
			_wcsupr_s( arg[i], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Unknown option: \"%s\"\r\n", arg[i] );
			return 1;
		}
		else if ( !GotPid && !FindName ) {
			sep1 = NULL;

			if ( ParseInt( arg[i], (unsigned int*) &Pid, &sep1 ) )
				FindName = arg[i];
			else
				GotPid = TRUE;

			if ( sep1 )
				if ( *sep1 )
					FindName = arg[i], Pid = 0;

			if ( FindName )
				GotPid = TRUE;
		}
		else {
			_wcsupr_s( arg[i], MAX_ARG_LEN );
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Syntax error: \"%s\"\r\n", arg[i] );
			return 1;
		}
	}


	if ( !GotPid )
		Pid = GetCurrentProcessId();


	if ( Pause )
		CL_SetPauseLines( Pause );

	CL_PickHighlightColor( NoHighlight );

	if ( NoHighlight )
		HighlightSpecs[0] = '\0';

	if ( !Width ) {
		if ( CL_ConsoleHandle( STD_OUTPUT_HANDLE ) ) {
			Width = CL_GetConsoleWidth( PROCTREE_WIDTH_DEFAULT + 1 ) - 1;

			if ( Width < PROCTREE_WIDTH_MIN )
				Width = PROCTREE_WIDTH_MIN;
			else if ( Width > PROCTREE_WIDTH_MAX )
				Width = PROCTREE_WIDTH_MAX;
		}
		else
			Width = PROCTREE_WIDTH_DEFAULT;
	}


	if ( rv = ShowAncestry( Pid, FindName, HighlightSpecs[0] ? HighlightSpecs : NULL, Width, Flags ) ) {

		if ( rv == USER_ABORT )
			return 3;
		else if ( rv == PROCTREE_ERROR_PID_NOT_FOUND )
			wsprintf( ErrorText, L"Process ID not found: 0x%04X (%u)", Pid, Pid );
		else if ( rv == PROCTREE_ERROR_NO_MATCH )
			wcscpy_s( ErrorText, SHORT_BUF_LEN, L"No process matching the given wildcards was found." );
		else
			GetErrorText( rv, ErrorText );

		DisplayErrorHeader();
		Qprintf( ERRHANDLE, L"%s\r\n", ErrorText );
		return 2;
	}


	return rv;
}



DLLExports int f_ancestry( LPTSTR String )
{
	if ( !String )
		return 1;


	wchar_t temp[SHORT_BUF_LEN]		= L"";

	unsigned int Flags			= 0;

	DWORD Pid					= 0;
	DWORD *Ancestors			= NULL;
	wchar_t *AncestorNames		= NULL;
	unsigned int NumAncestors	= 0;
	unsigned int NameLen		= 0;

	unsigned int Start			= 0;
	int Dir						= +1;

	size_t len					= 0;


	ParseArgs( String, PARSE_BREAK_COMMAS | PARSE_FORCE_UPPERCASE );


	if ( arglen[0] ) {

		if ( !wcscmp( arg[0], L"." ) )
			Pid = GetCurrentProcessId();
		else if ( ParseInt( arg[0], (unsigned int*) &Pid, NULL ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad PID value: \"%s\"\r\n", arg[0] );
			return -1;
		}

		if ( !Pid ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"PID must be nonzero: \"%s\"\r\n", arg[0] );
			return -1;
		}
	}

	if ( arglen[1] ) {
		if ( ParseInt( arg[1], &Flags, NULL ) ) {
			DisplayErrorHeader();
			Qprintf( ERRHANDLE, L"Bad flags value: \"%s\"\r\n", arg[1] );
			return -1;
		}
	}


	if ( !Pid )
		Pid = GetCurrentProcessId();


	String[0] = '\0';

	if ( !GetAncestors( Pid, &Ancestors, &AncestorNames, &NumAncestors, &NameLen ) )
		return 0;


	if ( ( Flags & 64 ) == 0 )
		Start = NumAncestors - 1, Dir = -1;

	for ( unsigned int i = Start; i < NumAncestors; i += Dir ) {
		switch ( Flags & 3 ) {

		case 1 :
			wsprintf( String + wcslen( String ), L"%X ", Ancestors[i] );
			break;

		case 2 :
			wsprintf( String + wcslen( String ), L"%u ", Ancestors[i] );
			break;

		case 3 :
			wcscpy_s( temp, SHORT_BUF_LEN, AncestorNames + ( i * NameLen ) );
			AddQuotes( temp );
			wcscat_s( String, MAX_ARG_LEN, temp );
			wcscat_s( String, MAX_ARG_LEN, L" " );
			break;

		default :
			wsprintf( String + wcslen( String ), L"0x%04X ", Ancestors[i] );
		}
	}


	len = wcslen( String );

	while ( len ) {
		if ( iswspace( String[len - 1] ) ) {
			String[len - 1] = '\0';
			len--;
		}
		else
			break;
	}


	if ( Flags & 128 )
		CharLowerW( String );


	free( Ancestors );
	free( AncestorNames );
	return 0;
}



const int NumHelpTexts							= 3;
const wchar_t HelpFeature[NumHelpTexts][11]		= { L"ancestry", L"proctree", L"@ancestry" };
const wchar_t *HelpText[NumHelpTexts]			= { AncestryHelpText, ProcTreeHelpText, F_AncestryHelpText };



#ifndef	NEWHELP_VERSION

LPTSTR UsageText( const wchar_t *inString )
{
	if ( !inString )
		return NULL;


	wchar_t *ch = wcschr( (wchar_t*) inString, '\n' );


	if ( !ch )
		return NULL;

	while ( *ch == '\n' )
		ch++;

	return ch;
}

#endif	//	NEWHELP_VERSION




DLLExports int WINAPI Usage( LPTSTR inString, LPTSTR outString )
{
	if ( ( inString[0] == '@' ) || ( inString[0] == '_' ) )
		return 0;

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			wcscpy_s( outString, SHORT_BUF_LEN, UsageText( HelpText[i] ) );
			return 1;
		}
	}

	return 0;
}


#ifndef NEWHELP_VERSION

//		Simpler MessageBox() help, useful for smaller plugins which don't merit a full HTML help system:

DLLExports int WINAPI Help( LPTSTR inString )
{
	wchar_t WindowTitle[SHORT_BUF_LEN]		= L"";


	wsprintf( WindowTitle, L"%s Plugin Help", PLUGIN_NAME );

	for ( int i = 0; i < NumHelpTexts; i++ ) {
		if ( !_wcsicmp( inString, HelpFeature[i] ) ) {
			MessageBox( GetForegroundWindow(), HelpText[i], WindowTitle, MB_OK );
			return 1;
		}
	}

	return 0;
}

#endif		//	NEWHELP_VERSION


/*

		0.99.5.0	2021-12-21
					Fixed a problem calling TakeCommandIPC() in 32-bit TCC.

		0.99.6.0	2021-12-22
					No longer link to TakeCommandIPC() dynamically.  That hasn't
					been necessary for a long, long time.

		1.0.0.0		2022-03-25
					Added /N? /O? /T? to give more info on those three options.
					Removed dependency on QueryFullProcessImageNameW() in hopes of
					making the plugin work under Windows XP.

		1.0.1.0		2022-03-25
					Replaced _DOSsify() with the more sophisticated FixKernelPath().

		1.0.2.0		2022-03-28
					Updated the help text and the error messages for /N /O /T to
					mention the existence of /N? /O? /T?

		1.0.3.0		2022-04-01
					Added /H: to highlight matching process names.

		1.0.4.0		2023-10-04
					Bug fix:  If a process's alleged parent was started after the process,
					then your daddy ain't your daddy (but your daddy don't know).

		1.0.4.1		2023-10-04
					Formatting change:  /I now uses a right guillemet instead of an
					arrowhead to mark 32-bit processes.  Added and removed spaces
					in the /I info to make things line up a bit better.

		1.0.5.0		2023-10-05
					/H: highlights matching window titles / service names as well as
					process names.  /HL: is a synonym for /H:

		1.0.5.1		2023-10-12
					Updated the web address in the PLUGININFO structure.

		1.0.5.2		2023-10-15
					Tweaked ShowCmdHelp() to include VER_PATCH.

		1.1.0.0		2024-05-03
					Added ANCESTRY and @ANCESTRY.  Allow . as a synonym for %_PID.

		1.1.1.0		2024-05-08
					Rewrote ANCESTRY to recycle PROCTREE's display code.  Now many
					PROCTREE options work in ANCESTRY.

*/
